#!/usr/bin/env python

"""
Tracing tools built on USDT
"""

from usdt import Probe, Provider
import functools


class fbt(object):
    """
    simple function boundary tracing decorator
    """

    def __init__(self, func):
        self.func = func
        probename = func.__name__
        self.entry_probe = Probe(probename, "entry", ["char *"])
        self.return_probe = Probe(probename, "return", ["char *"])
        self.provider = Provider("python-fbt", "fbt")
        self.provider.add_probe(self.entry_probe)
        self.provider.add_probe(self.return_probe)
        self.provider.enable()

    def __call__(self, *args):
        self.entry_probe.fire([", ".join([str(x) for x in args])])
        ret = self.func(*args)
        self.return_probe.fire([str(ret)])
        return ret

    def __get__(self, obj, objtype):
        """Support instance methods. (http://stackoverflow.com/a/3296318)"""
        return functools.partial(self.__call__, obj)



#!/usr/bin/env python

"""
Tools for logging that leverage USDT
"""

from usdt import Probe, Provider
import logging

_LEVELS = [logging.CRITICAL, logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG, logging.NOTSET]
_DESC = {
    logging.CRITICAL: "critical",
    logging.ERROR: "error",
    logging.WARNING: "warning",
    logging.INFO: "info",
    logging.DEBUG: "debug",
    logging.NOTSET: "notset",
}


class DtraceHandler(logging.Handler):
    """ Handler to fire USDT probes with log messages """

    def __init__(self):
        logging.Handler.__init__(self)

        self.provider = Provider("python", "dtrace-logger")

        self.probes = {}
        for key in _DESC.keys():
            self.probes[key] = Probe("logging", _DESC[key], ["int", "char *"])
            self.provider.add_probe(self.probes[key])
        self.provider.enable()

    def emit(self, record):
        """ Fire the appropriate USDT probe for this record's log level """
        probe = None
        for level in _LEVELS:
            if record.levelno >= level:
                probe = self.probes[level]
                break
        if not probe:
            probe = self.probes[logging.NOTSET]
        probe.fire([record.levelno, self.format(record)])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_usdt
----------------------------------

Tests for `usdt` module.
"""

import os
import sys
import logging
import unittest
import subprocess

import usdt

from usdt.tracer import fbt
from usdt.log import DtraceHandler

if not os.geteuid() == 0:
    sys.exit('Please run with sudo (for launching DTrace).')

class DtraceTest(unittest.TestCase):
    def setUp(self):
        begin_probe = ':::BEGIN { printf("initialized\\n"); }'

        cmd = ['dtrace', '-Z', '-q', '-n', begin_probe]
        for glob in self.dtrace_globs:
            cmd += ['-n', glob]
        self.dtrace = subprocess.Popen(cmd, stdout=subprocess.PIPE)

        # wait for dtrace to print first line of output (based on the :::BEGIN
        #  probe) to indicate that it's readuy
        self.assertTrue(self.get_next_message().endswith('initialized\n'))

    def get_next_message(self):
        return self.dtrace.stdout.readline()

    def tearDown(self):
        self.dtrace.terminate()


class ProbeTest(DtraceTest):
    dtrace_globs = [
        '::hello:int { printf("%d\\n", arg1); }',
        '::hello:name { printf("%s\\n", copyinstr(arg0)); }'
    ]

    def test_probe(self):
        test_prov = usdt.Provider("python", "provmod")
        test_probe = usdt.Probe("hello", "name", ["char *"])
        test_prov.add_probe(test_probe)
        test_prov.enable()

        test_probe.fire(["Hello World"])
        self.assertEqual(self.get_next_message(), 'Hello World\n')

    def test_probe_2(self):
        test_prov = usdt.Provider("python", "provmod")
        int_probe = usdt.Probe("hello", "int", ["char *", "int"])
        test_prov.add_probe(int_probe)
        test_prov.enable()

        int_probe.fire(["Number Test", 5])
        self.assertEqual(self.get_next_message(), '5\n')


class FunctionBoundaryTracerTest(DtraceTest):
    dtrace_globs = [
        ':fbt:: { printf("%s:%s:%s:%s %s\\n", probeprov, probemod, probefunc, probename, copyinstr(arg0)); }'
    ]

    def test_fbt(self):
        @fbt
        def hello(arg1, arg2):
            return True

        hello(1, 2)
        self.assertTrue(self.get_next_message().endswith('hello:entry 1, 2\n'))
        self.assertTrue(self.get_next_message().endswith('hello:return True\n'))

    @fbt
    def hello2(self, arg1, arg2):
        return True

    def test_fbt_2(self):
        """Same as test_fbt, but tracing a class method."""
        self.hello2(1, 2)
        the_message = self.get_next_message()
        self.assertTrue('hello2:entry test_fbt_2 (' in the_message)
        self.assertTrue(the_message.endswith('FunctionBoundaryTracerTest), 1, 2\n'))
        self.assertTrue(self.get_next_message().endswith('hello2:return True\n'))


class LoggingTest(DtraceTest):
    dtrace_globs = [
        '::logging: { printf("%s:%s:%s:%s %d %s\\n", probeprov, probemod, probefunc, probename, arg0, copyinstr(arg1)); }'
    ]

    def get_probename(self, logline):
        probeprov, probemod, probefunc, more = logline.split(':')
        probename, arg0, arg1 = more.split(' ', 2)
        return probename

    def test_logging(self):
        logger = logging.getLogger()
        logger.setLevel(logging.NOTSET)

        dtrace_handler = DtraceHandler()
        dtrace_handler.setLevel(logging.NOTSET)

        logger.addHandler(dtrace_handler)

        logger.debug("debug message")
        logger.info("info message")
        logger.warning("warning message")
        logger.error("error message")
        logger.critical("critical message")
        logger.log(logging.NOTSET + 1, "notset message")

        self.assertEqual(self.get_probename(self.get_next_message()), 'debug')
        self.assertEqual(self.get_probename(self.get_next_message()), 'info')
        self.assertEqual(self.get_probename(self.get_next_message()), 'warning')
        self.assertEqual(self.get_probename(self.get_next_message()), 'error')
        self.assertEqual(self.get_probename(self.get_next_message()), 'critical')
        self.assertEqual(self.get_probename(self.get_next_message()), 'notset')


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

"""
libusdt bindings for Python
"""

from __future__ import print_function
import os
from ctypes import cdll, c_char_p, c_int, c_void_p, cast, POINTER

__author__ = 'Nahum Shalman'
__email__ = 'nshalman@omniti.com'
__version__ = '0.1.3'

HAVE_USDT = False
FAKE_DTRACE = False

try:
    HERE = os.path.dirname(os.path.realpath(__file__))
    _LIBUSDT = cdll.LoadLibrary(HERE + "/libusdt.so")
    HAVE_USDT = True
except OSError:
    # Didn't find the library, probably not supported
    pass

if HAVE_USDT:
    _LIBUSDT.usdt_create_provider.argtypes = [c_char_p, c_char_p]
    _LIBUSDT.usdt_create_probe.argtypes = [c_char_p, c_char_p, c_int, c_void_p]

    _LIBUSDT.usdt_create_provider.restype = POINTER(c_void_p)
    _LIBUSDT.usdt_create_probe.restype = POINTER(c_void_p)

    class Probe(object):
        """ a USDT probe """

        def __init__(self, func, name, arg_desc):
            self._LIBUSDT = _LIBUSDT
            self.length = len(arg_desc)
            args = (c_char_p * self.length)()
            for i in range(self.length):
                args[i] = arg_desc[i]
            self.probedef = self._LIBUSDT.usdt_create_probe(func,
                                                            name, self.length, args)

        def fire(self, args):
            """ fire the probe """
            if len(args) == self.length:
                c_args = (c_void_p * self.length)()
                for i in range(self.length):
                    c_args[i] = cast(args[i], c_void_p)
                self._LIBUSDT.usdt_fire_probedef(self.probedef, self.length, c_args)

        def __del__(self):
            self._LIBUSDT.usdt_probe_release(self.probedef)
            del self._LIBUSDT
            del self.length
            del self.probedef

    class Provider(object):
        """ a USDT provider """

        def __init__(self, provider="python-dtrace", module="default_module"):
            self._LIBUSDT = _LIBUSDT
            self.provider = self._LIBUSDT.usdt_create_provider(provider, module)
            self.probes = []

        def add_probe(self, probe):
            """ add a probe to this provider """
            self.probes.append(probe)
            self._LIBUSDT.usdt_provider_add_probe(self.provider, probe.probedef)

        def enable(self):
            """ enable the provider """
            return(self._LIBUSDT.usdt_provider_enable(self.provider))

        def __del__(self):
            for probe in self.probes:
                del probe
            del self.probes
            self._LIBUSDT.usdt_provider_disable(self.provider)
            self._LIBUSDT.usdt_provider_free(self.provider)
            del self.provider
            del self._LIBUSDT
else:
    from sys import stderr

    class Probe(object):
        """ a fake USDT probe """

        def __init__(self, name, func, arg_desc):
            self.name = name
            self.func = func
            self.provider = None
            self.arg_desc = arg_desc

        def fire(self, args):
            """ send probe info to stderr if requested """
            if FAKE_DTRACE and self.provider and self.provider.enabled:
                print(self.provider.provider, self.provider.module,
                      self.name, self.func, args,
                      file=stderr)

        def __del__(self):
            pass

    class Provider(object):
        """ a fake USDT provider """
        probes = []

        def __init__(self, provider="python-dtrace", module="default_module"):
            self.provider = provider
            self.module = module
            self.enabled = False

        def add_probe(self, probe):
            """ add a probe to this provider """
            self.probes.append(probe)
            probe.provider = self

        def enable(self):
            """ enable this (fake) provider """
            self.enabled = True

        def __del__(self):
            pass



import os
import platform
import subprocess

OS = platform.os.uname()[0]
ARCH = platform.architecture()[0]


def post_build():
    """Compile libusdt"""
    extra = ""
    if OS == "SunOS" and ARCH == "32bit":
        extra = "ARCH=i386"

    linker_flag = "--whole-archive"
    compiler = subprocess.check_output(["cc", "--version"])
    if b"clang" in compiler:
        linker_flag = "-force_load"

    libdir = os.getcwd() + "/build/lib/usdt/libusdt"
    source = libdir + "/libusdt.a"
    library = libdir + ".so"
    try:
        if os.system("cd %s ; make %s clean all" % (libdir, extra)) == 0:
            os.system("gcc -g -shared -o %s -Wl,%s %s" %
                      (library, linker_flag, source))
    except:
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys

from setuptools import setup, find_packages
from setuptools.command.build_py import build_py as _build_py

readme = open('README.rst').read()
history = open('HISTORY.rst').read().replace('.. :changelog:', '')

requirements = [
    # TODO: put package requirements here
]

test_requirements = [
    # TODO: put package test requirements here
]

class build_py_(_build_py):
    def run(self):
        ret = _build_py.run(self)
        import site
        site.addsitedir(self.build_lib)
        sys.path.insert(0, self.build_lib)
        from usdt import build_hooks
        build_hooks.post_build()
        return ret

setup(
    name='usdt',
    version='0.1.3',
    description='Python libusdt bindings',
    long_description=readme + '\n\n' + history,
    author='Nahum Shalman',
    author_email='nshalman@omniti.com',
    url='https://github.com/nshalman/python-usdt',
    packages=[
        'usdt',
    ],
    package_dir={'usdt':
                 'usdt'},
    include_package_data=True,
    install_requires=requirements,
    license="BSD",
    zip_safe=False,
    keywords='usdt',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    cmdclass={'build_py': build_py_}
)
